#!/usr/bin/env python

# Written by John Hoffman
# see LICENSE.txt for license information

from BitTornado import PSYCO
if PSYCO.psyco:
    try:
        import psyco
        assert psyco.__version__ >= 0x010100f0
        psyco.full()
    except:
        pass

from download_bt1 import BT1Download
from RawServer import RawServer
from RateLimiter import RateLimiter
from ServerPortHandler import MultiHandler
from parsedir import parsedir
from random import seed
from socket import error as socketerror
from threading import Event
from sys import argv, exit
import sys, os
from time import time
from __init__ import createPeerID, mapbase64, version
from threading import Event

try:
    True
except:
    True = 1
    False = 0


def fmttime(n):
    if n is None:
        return 'downloading'
    n = int(n)
    m, s = divmod(n, 60)
    h, m = divmod(m, 60)
    if h > 1000000:
        return 'downloading'
    return '%d:%02d:%02d' % (h, m, s)

class SingleDownload:
    def __init__(self, controller, hash, response, config, myid):
        self.controller = controller
        self.hash = hash
        self.response = response
        self.config = config
        
        self.doneflag = Event()
        self.waiting = True
        self.checking = False
        self.working = False
        self.seed = False

        self.status_msg = ''
        self.status_err = ['']
        self.status_errtime = 0
        self.status_done = 0.0

        self.rawserver = controller.handler.newRawServer(hash, self.doneflag)

        d = BT1Download(self.display, self.finished, self.error,
                        controller.exchandler, self.doneflag, config, response,
                        hash, myid, self.rawserver, controller.listen_port)
        self.d = d

        if not d.saveAs(self.saveAs):
            self.died()
            return
        self._hashcheckfunc = d.initFiles()
        if not self._hashcheckfunc:
            self.died()
            return

        controller.hashchecksched(self)

    def saveAs(self, name, length, saveas, isdir):
        return self.controller.saveAs(self.hash, name, saveas, isdir)

    def hashcheck_start(self, donefunc):
        if self._is_dead():
            return
        self.waiting = False
        self.checking = True
        self._hashcheckfunc(donefunc)

    def hashcheck_callback(self):
        self.checking = False
        if self._is_dead():
            return
        if not self.d.startEngine(ratelimiter = self.controller.ratelimiter):
            self.died()
            return
        self.d.startRerequester()
        self.statsfunc = self.d.startStats()
        self.rawserver.start_listening(self.d.getPortHandler())
        self.working = True


    def died(self):
        self.doneflag.set()
        self.rawserver.shutdown()
        self.waiting = False
        self.checking = False
        self.working = False
        self.controller.died(self.hash)

    def is_dead(self):
        return self.doneflag.isSet()

    def _is_dead(self):
        if self.is_dead():
            self.died()
            return True
        return False

    def shutdown(self):
        if not self.is_dead():
            if self.working:
                self.d.shutdown()
            self.died()


    def display(self, activity = None, fractionDone = None):
        # really only used by StorageWrapper now
        if activity:
            self.status_msg = activity
        if fractionDone is not None:
            self.status_done = float(fractionDone)

    def finished(self):
        self.seed = True

    def error(self, msg):
        self.status_err.append(msg)
        self.status_errtime = time()


class LaunchMany:
    def __init__(self, config, Output):
        self.config = config
        self.Output = Output

        self.torrent_dir = config['torrent_dir']
        self.torrent_cache = {}
        self.file_cache = {}
        self.blocked_files = {}
        self.scan_period = config['parse_dir_interval']
        self.stats_period = config['display_interval']

        self.torrent_list = []
        self.downloads = {}
        self.counter = 0
        self.doneflag = Event()

        self.hashcheck_queue = []
        self.hashcheck_current = None
        
        self.rawserver = RawServer(self.doneflag, config['timeout_check_interval'],
                          config['timeout'], ipv6_enable = config['ipv6_enabled'],
                          failfunc = self.failed, errorfunc = self.exchandler)
        try:
            self.listen_port = self.rawserver.find_and_bind(
                            config['minport'], config['maxport'], config['bind'],
                            ipv6_socket_style = config['ipv6_binds_v4'],
                            upnp = config['upnp_nat_access'])
        except socketerror, e:
            self.failed("Couldn't listen - " + str(e))
            return

        self.ratelimiter = RateLimiter(self.rawserver.add_task,
                                       config['upload_unit_size'])
        self.ratelimiter.set_upload_rate(config['max_upload_rate'])

        self.handler = MultiHandler(self.rawserver, self.doneflag)
        seed(createPeerID())
        self.rawserver.add_task(self.scan, 0)
        self.rawserver.add_task(self.stats, 0)

        self.handler.listen_forever()

        for hash in self.torrent_list:
            self.Output.message('dropped "'+self.torrent_cache[hash]['path']+'"')
            self.downloads[hash].shutdown()


    def scan(self):
        self.rawserver.add_task(self.scan, self.scan_period)
                                
        r = parsedir(self.torrent_dir, self.torrent_cache,
                     self.file_cache, self.blocked_files,
                     return_metainfo = True, errfunc = self.Output.message)

        ( self.torrent_cache, self.file_cache, self.blocked_files,
            added, removed ) = r

        for hash, data in removed.items():
            self.Output.message('dropped "'+data['path']+'"')
            self.remove(hash)
        for hash, data in added.items():
            self.Output.message('added "'+data['path']+'"')
            self.add(hash, data)

    def stats(self):            
        self.rawserver.add_task(self.stats, self.stats_period)
        data = []
        for hash in self.torrent_list:
            cache = self.torrent_cache[hash]
            name = cache['path']
            size = cache['length']
            d = self.downloads[hash]
            progress = '0.0%'
            peers = 0
            seeds = 0
            seedsmsg = "S"
            uprate = 0.0
            dnrate = 0.0
            upamt = 0
            dnamt = 0
            t = 0
            if d.is_dead():
                status = 'stopped'
            elif d.waiting:
                status = 'waiting for hash check'
            elif d.checking:
                status = d.status_msg
                progress = '%.1f%%' % (d.status_done*100)
            else:
                stats = d.statsfunc()
                s = stats['stats']
                if d.seed:
                    status = 'seeding'
                    progress = '100.0%'
                    seeds = s.numOldSeeds
                    seedsmsg = "s"
                else:
                    if s.numSeeds + s.numPeers:
                        t = stats['time']
                        if t == 0:  # unlikely
                            t = 0.01
                        status = fmttime(t)
                    else:
                        t = -1
                        status = 'connecting to peers'
                    progress = '%.1f%%' % (int(stats['frac']*1000)/10.0)
                    seeds = s.numSeeds
                    dnrate = stats['down']
                peers = s.numPeers
                uprate = stats['up']
                upamt = s.upTotal
                dnamt = s.downTotal
                   
            if d.is_dead() or d.status_errtime+300 > time():
                msg = d.status_err[-1]
            else:
                msg = ''

            data.append(( name, status, progress, peers, seeds, seedsmsg,
              uprate, dnrate, upamt, dnamt, size, t ))
        self.Output.display(data)

    def remove(self, hash):
        self.downloads[hash].shutdown()
        del self.downloads[hash]
        self.torrent_list.remove(hash)
        
    def add(self, hash, data):
        c = self.counter
        self.counter += 1
        x = ''
        for i in xrange(3):
            x = mapbase64[c & 0x3F]+x
            c >>= 6
        peer_id = createPeerID(x)
        d = SingleDownload(self, hash, data['metainfo'], self.config, peer_id)
        self.torrent_list.append(hash)
        self.downloads[hash] = d


    def saveAs(self, hash, name, saveas, isdir):
        x = self.torrent_cache[hash]
        if saveas:
            saveas = os.path.join(saveas,x['file'][:-1-len(x['type'])])
        else:
            saveas = x['path'][:-1-len(x['type'])]
        if isdir and not os.path.isdir(saveas):
            try:
                os.mkdir(saveas)
            except:
                raise OSError("couldn't create directory for "+x.path)
        return saveas

    def hashchecksched(self, download):
        self.hashcheck_queue.append(download)
        if not self.hashcheck_current:
            self._hashcheck_start()

    def _hashcheck_start(self):
        self.hashcheck_current = self.hashcheck_queue.pop(0)
        self.hashcheck_current.hashcheck_start(self.hashcheck_callback)

    def hashcheck_callback(self):
        self.hashcheck_current.hashcheck_callback()
        if self.hashcheck_queue:
            self._hashcheck_start()
        else:
            self.hashcheck_current = None

    def died(self, hash):
        pass

    def failed(self, s):
        self.Output.message('FAILURE: '+s)

    def exchandler(self, s):
        print 'EXCEPTION'
        print s
